[comment {-*- tcl -*- doctools = format-nopackage.inc}]
[para]

The package [package img::[vset FORMATNAME]] is [emph not] loaded when doing a
[cmd {package require Img}], but must be loaded explicitly via
[cmd "package require img::[vset FORMATNAME]"].

[para]
