[comment {-*- tcl -*- doctools = format-footer.inc}]

[see_also img img-bmp img-dted img-flir img-gif img-ico img-jpeg img-pcx img-pixmap img-png img-ppm img-ps img-raw img-sgi img-sun img-tga img-tiff img-window img-xbm img-xpm]
[manpage_end]
